@@ -1,6 +1,6 @@
 ;;; autoload.el --- maintain autoloads in loaddefs.el.
 
-;;; Copyright (C) 1991, 1992, 1993 Free Software Foundation, Inc.
+;;; Copyright (C) 1991, 1992, 1993, 1994 Free Software Foundation, Inc.
 ;;;
 ;; Author: Roland McGrath <roland@gnu.ai.mit.edu>
 ;; Keywords: maint
@@ -21,9 +21,9 @@
 ;;; 02139, USA.
 ;;;
 
-;;; Commentary;:
+;;; Commentary:
 
-;; This code helps GNU Emacs maintainers keep the autoload.el file up to
+;; This code helps GNU Emacs maintainers keep the loaddefs.el file up to
 ;; date.  It interprets magic cookies of the form ";;;###autoload" in
 ;; lisp source files in various useful ways.  To learn more, read the
 ;; source; if you're going to use this, you'd better be able to.
@@ -95,6 +95,16 @@ the section of autoloads for a file.")
 (put 'defconst 'doc-string-elt 3)
 (put 'defmacro 'doc-string-elt 3)
 
+(defun autoload-trim-file-name (file)
+  ;; Returns a relative pathname of FILE
+  ;; starting from the directory that loaddefs.el is in.
+  ;; That is normally a directory in load-path,
+  ;; which means Emacs will be able to find FILE when it looks.
+  ;; Any extra directory names here would prevent finding the file.
+  (setq file (expand-file-name file))
+  (file-relative-name file
+		      (file-name-directory generated-autoload-file)))
+
 (defun generate-file-autoloads (file)
   "Insert at point a loaddefs autoload section for FILE.
 autoloads are generated for defuns and defmacros in FILE
@@ -109,7 +119,8 @@ are used."
 			 (substring name 0 (match-beginning 0))
 		       name)))
 	(print-length nil)
-	(float-output-format "%.20e")
+	(print-readably t)		; This does something in Lucid Emacs.
+	(float-output-format nil)
 	(done-any nil)
 	(visited (get-file-buffer file))
 	output-end)
@@ -149,7 +160,7 @@ are used."
 		    (if (eolp)
 			;; Read the next form and make an autoload.
 			(let* ((form (prog1 (read (current-buffer))
-				       (forward-line 1)))
+				       (or (bolp) (forward-line 1))))
 			       (autoload (make-autoload form load-name))
 			       (doc-string-elt (get (car-safe form)
 						    'doc-string-elt)))
@@ -173,28 +184,41 @@ are used."
 						      (princ " " outbuf)))
 					  autoload))
 				(princ "\"\\\n" outbuf)
-				(princ (substring
-					(prin1-to-string (car elt)) 1)
-				       outbuf)
-				(if (null (cdr elt))
-				    (princ ")" outbuf)
-				  (princ " " outbuf)
+				(let ((begin (save-excursion
+					       (set-buffer outbuf)
+					       (point))))
 				  (princ (substring
-					  (prin1-to-string (cdr elt))
-					  1)
-					 outbuf))
-				(terpri outbuf))
-			    (print autoload outbuf)))
-		      ;; Copy the rest of the line to the output.
-		      (let ((begin (point)))
-			(forward-line 1)
-			(princ (buffer-substring begin (point)) outbuf))))
-			((looking-at ";")
-			 ;; Don't read the comment.
-			 (forward-line 1))
-			(t
-			 (forward-sexp 1)
-			 (forward-line 1)))))))
+					  (prin1-to-string (car elt)) 1)
+					 outbuf)
+				  ;; Insert a backslash before each ( that
+				  ;; appears at the beginning of a line in
+				  ;; the doc string.
+				  (save-excursion
+				    (set-buffer outbuf)
+				    (save-excursion
+				      (while (search-backward "\n(" begin t)
+					(forward-char 1)
+					(insert "\\"))))
+				  (if (null (cdr elt))
+				      (princ ")" outbuf)
+				    (princ " " outbuf)
+				    (princ (substring
+					    (prin1-to-string (cdr elt))
+					    1)
+					   outbuf))
+				  (terpri outbuf)))
+			    (let ((print-escape-newlines t))
+			      (print autoload outbuf))))
+			  ;; Copy the rest of the line to the output.
+			  (let ((begin (point)))
+			    (forward-line 1)
+			    (princ (buffer-substring begin (point)) outbuf))))
+		   ((looking-at ";")
+		    ;; Don't read the comment.
+		    (forward-line 1))
+		   (t
+		    (forward-sexp 1)
+		    (forward-line 1)))))))
 	(or visited
 	    ;; We created this buffer, so we should kill it.
 	    (kill-buffer (current-buffer)))
@@ -203,11 +227,24 @@ are used."
     (if done-any
 	(progn
 	  (insert generate-autoload-section-header)
-	  (prin1 (list 'autoloads autoloads-done load-name file
+	  (prin1 (list 'autoloads autoloads-done load-name
+		       (autoload-trim-file-name file)
 		       (nth 5 (file-attributes file)))
 		 outbuf)
 	  (terpri outbuf)
-	  (insert ";;; Generated autoloads from " file "\n")
+	  (insert ";;; Generated autoloads from "
+		  (autoload-trim-file-name file) "\n")
+	  ;; Warn if we put a line in loaddefs.el
+	  ;; that is long enough to cause trouble.
+	  (while (< (point) output-end)
+	    (let ((beg (point)))
+	      (end-of-line)
+	      (if (> (- (point) beg) 900)
+		  (progn
+		    (message "A line is too long--over 900 characters")
+		    (sleep-for 2)
+		    (goto-char output-end))))
+	    (forward-line 1))
 	  (goto-char output-end)
 	  (insert generate-autoload-section-trailer)))
     (message "Generating autoloads for %s...done" file)))
@@ -226,7 +263,7 @@ autoloads go somewhere else.")
 		     (if (string-match "\\.elc?$" name)
 			 (substring name 0 (match-beginning 0))
 		       name)))
-	(done nil)
+	(found nil)
 	(existing-buffer (get-file-buffer file)))
     (save-excursion
       ;; We want to get a value for generated-autoload-file from
@@ -237,52 +274,56 @@ autoloads go somewhere else.")
 	(save-restriction
 	  (widen)
 	  (goto-char (point-min))
-	  (while (search-forward generate-autoload-section-header nil t)
+	  ;; Look for the section for LOAD-NAME.
+	  (while (and (not found)
+		      (search-forward generate-autoload-section-header nil t))
 	    (let ((form (condition-case ()
 			    (read (current-buffer))
 			  (end-of-file nil))))
-	      (if (string= (nth 2 form) load-name)
-		  (let ((begin (match-beginning 0))
-			(last-time (nth 4 form))
-			(file-time (nth 5 (file-attributes file))))
-		    (if (and (or (null existing-buffer)
-				 (not (buffer-modified-p existing-buffer)))
-			     (listp last-time) (= (length last-time) 2)
-			     (or (> (car last-time) (car file-time))
-				 (and (= (car last-time) (car file-time))
-				      (>= (nth 1 last-time)
-					  (nth 1 file-time)))))
-			(message "Autoload section for %s is up to date."
-				 file)
-		      (search-forward generate-autoload-section-trailer)
-		      (delete-region begin (point))
-		      (generate-file-autoloads file))
-		    (setq done t))))))
-	(if done
-	    ;; There was an existing section and we have updated it.
-	    ()
-	  (if (save-excursion
-		(set-buffer (find-file-noselect file))
-		(save-excursion
-		  (save-restriction
-		    (widen)
-		    (goto-char (point-min))
-		    (search-forward generate-autoload-cookie nil t))))
-	      ;; There are autoload cookies in FILE.
-	      ;; Have the user tell us where to put the new section.
-	      (progn
-		(save-window-excursion
-		(switch-to-buffer (current-buffer))
-		(with-output-to-temp-buffer "*Help*"
-		  (princ (substitute-command-keys
-			  (format "\
-Move point to where the autoload section
-for %s should be inserted.
-Then do \\[exit-recursive-edit]."
-				  file))))
-		(recursive-edit)
-		(beginning-of-line))
-		(generate-file-autoloads file)))))
+	      (cond ((string= (nth 2 form) load-name)
+		     ;; We found the section for this file.
+		     ;; Check if it is up to date.
+		     (let ((begin (match-beginning 0))
+			   (last-time (nth 4 form))
+			   (file-time (nth 5 (file-attributes file))))
+		       (if (and (or (null existing-buffer)
+				    (not (buffer-modified-p existing-buffer)))
+				(listp last-time) (= (length last-time) 2)
+				(or (> (car last-time) (car file-time))
+				    (and (= (car last-time) (car file-time))
+					 (>= (nth 1 last-time)
+					     (nth 1 file-time)))))
+			   (progn
+			     (message "Autoload section for %s is up to date."
+				      file)
+			     (setq found 'up-to-date))
+			 (search-forward generate-autoload-section-trailer)
+			 (delete-region begin (point))
+			 (setq found t))))
+		    ((string< load-name (nth 2 form))
+		     ;; We've come to a section alphabetically later than
+		     ;; LOAD-NAME.  We assume the file is in order and so
+		     ;; there must be no section for LOAD-NAME.  We will
+		     ;; insert one before the section here.
+		     (goto-char (match-beginning 0))
+		     (setq found 'new)))))
+	  (or (eq found 'up-to-date)
+	      (and (eq found 'new)
+		   ;; Check that FILE has any cookies before generating a
+		   ;; new section for it.
+		   (save-excursion
+		     (set-buffer (find-file-noselect file))
+		     (save-excursion
+		       (widen)
+		       (goto-char (point-min))
+		       (if (search-forward (concat "\n"
+						   generate-autoload-cookie)
+					   nil t)
+			   nil
+			 (if (interactive-p)
+			     (message file " has no autoloads"))
+			 t))))
+	      (generate-file-autoloads file))))
       (if (interactive-p) (save-buffer))
       (if (and (null existing-buffer)
 	       (setq existing-buffer (get-file-buffer file)))
@@ -305,12 +346,11 @@ Update sections of the current buffer generated by \\[update-file-autoloads]."
 		   (or (get-file-buffer file)
 		       (file-exists-p file)))
 	      ()
-	    (setq file (if (y-or-n-p (format "Library \"%s\" (load \
-file \"%s\") doesn't exist.  Remove its autoload section? "
+	    (setq file (if (y-or-n-p (format "Can't find library `%s'; remove its autoloads? "
 					     (nth 2 form) file))
 			   t
 			 (condition-case ()
-			     (read-file-name (format "Find \"%s\" load file: "
+			     (read-file-name (format "Find `%s' load file: "
 						     (nth 2 form))
 					     nil nil t)
 			   (quit nil)))))
@@ -325,8 +365,9 @@ file \"%s\") doesn't exist.  Remove its autoload section? "
 (defun update-directory-autoloads (dir)
   "Run \\[update-file-autoloads] on each .el file in DIR."
   (interactive "DUpdate autoloads for directory: ")
-  (mapcar 'update-file-autoloads
-	  (directory-files dir nil "\\.el$"))
+  (let ((enable-local-eval nil))
+    (mapcar 'update-file-autoloads
+	    (directory-files dir t "^[^=].*\\.el$")))
   (if (interactive-p)
       (save-excursion
 	(set-buffer (find-file-noselect generated-autoload-file))
@@ -338,22 +379,34 @@ file \"%s\") doesn't exist.  Remove its autoload section? "
 Runs \\[update-file-autoloads] on files and \\[update-directory-autoloads]
 on directories.  Must be used only with -batch, and kills Emacs on completion.
 Each file will be processed even if an error occurred previously.
-For example, invoke \"emacs -batch -f batch-update-autoloads *.el\""
+For example, invoke `emacs -batch -f batch-update-autoloads *.el'."
   (if (not noninteractive)
       (error "batch-update-autoloads is to be used only with -batch"))
   (let ((lost nil)
-	(args command-line-args-left))
-    (while args
-      (catch 'file
-	(condition-case lossage
-	    (if (file-directory-p (expand-file-name (car args)))
-		(update-directory-autoloads (car args))
-	      (update-file-autoloads (car args)))
-	  (error (progn (message ">>Error processing %s: %s"
-				 (car args) lossage)
-			(setq lost t)
-			(throw 'file nil)))))
-      (setq args (cdr args)))
+	(args command-line-args-left)
+	(enable-local-eval nil))	;Don't query in batch mode.
+    (message "Updating autoloads in %s..." generated-autoload-file)
+    (let ((frob (function
+ 		 (lambda (file)
+ 		   (condition-case lossage
+ 		       (update-file-autoloads file)
+ 		     (error
+ 		      (princ ">>Error processing ")
+ 		      (princ file)
+ 		      (princ ": ")
+ 		      (if (fboundp 'display-error)
+ 			  (display-error lossage nil)
+ 			(prin1 lossage))
+ 		      (princ "\n")
+ 		      (setq lost t)))))))
+      (while args
+ 	(if (file-directory-p (expand-file-name (car args)))
+ 	    (let ((rest (directory-files (car args) t "\\.el$")))
+ 	      (while rest
+ 		(funcall frob (car rest))
+ 		(setq rest (cdr rest))))
+ 	  (funcall frob (car args)))
+ 	(setq args (cdr args))))
     (save-some-buffers t)
     (message "Done")
     (kill-emacs (if lost 1 0))))
