@@ -1,8 +1,9 @@
 ;;; perl-mode.el --- Perl code editing commands for GNU Emacs
 
-;; Copyright (C) 1990  Free Software Foundation, Inc.
+;; Copyright (C) 1990, 1994  Free Software Foundation, Inc.
 
 ;; Author: William F. Mann
+;; Maintainer: FSF
 ;; Adapted-By: ESR
 ;; Keywords: languages
 
@@ -33,7 +34,7 @@
 ;; With argments to perl:
 ;; #!/usr/bin/perl -P-	 # -*-Perl-*-
 ;; To handle files included with do 'filename.pl';, add something like
-;; (setq auto-mode-alist (append (list (cons "\\.pl$" 'perl-mode))
+;; (setq auto-mode-alist (append (list (cons "\\.pl\\'" 'perl-mode))
 ;;                               auto-mode-alist))
 ;; to your .emacs file; otherwise the .pl suffix defaults to prolog-mode.
 
@@ -64,7 +65,7 @@
 ;; Known problems (these are all caused by limitations in the Emacs Lisp
 ;; parsing routine (parse-partial-sexp), which was not designed for such
 ;; a rich language; writing a more suitable parser would be a big job):
-;; 1)  Regular expression delimitors do not act as quotes, so special
+;; 1)  Regular expression delimiters do not act as quotes, so special
 ;;       characters such as `'"#:;[](){} may need to be backslashed
 ;;       in regular expressions and in both parts of s/// and tr///.
 ;; 2)  The globbing syntax <pattern> is not recognized, so special
@@ -146,32 +147,56 @@ The expansion is entirely correct because it uses the C preprocessor."
   (modify-syntax-entry ?| "." perl-mode-syntax-table)
 )
 
-(defconst perl-indent-level 4
+(defvar perl-font-lock-keywords
+  (list
+;   ("if" "until" "while" "elsif" "else" "unless" "for" "foreach" "continue"
+;    "exit" "die" "last" "goto" "next" "redo" "return" "local" "exec")
+   (concat "\\<\\("
+	   "continue\\|die\\|e\\(ls\\(e\\|if\\)\\|x\\(ec\\|it\\)\\)\\|"
+	   "for\\(\\|each\\)\\|goto\\|if\\|l\\(ast\\|ocal\\)\\|next\\|"
+	   "re\\(do\\|turn\\)\\|un\\(less\\|til\\)\\|while"
+	   "\\)\\>")
+;   ("#endif" "#else" "#ifdef" "#ifndef" "#if" "#include" "#define" "#undef")
+   (cons (concat "#\\(define\\|e\\(lse\\|ndif\\)\\|"
+		 "i\\(f\\(\\|def\\|ndef\\)\\|nclude\\)\\|undef\\)\\>")
+	 'font-lock-reference-face)
+   '("^[ \n\t]*sub[ \t]+\\([^ \t{]+\\)[ \t]*[{]" 1 font-lock-function-name-face)
+   '("[ \n\t{]*\\(eval\\)[ \n\t(;]" 1 font-lock-function-name-face)
+   '("\\(--- .* ---\\|=== .* ===\\)" . font-lock-string-face)
+   )
+  "Additional expressions to highlight in Perl mode.")
+
+(defvar perl-indent-level 4
   "*Indentation of Perl statements with respect to containing block.")
-(defconst perl-continued-statement-offset 4
+(defvar perl-continued-statement-offset 4
   "*Extra indent for lines not starting new statements.")
-(defconst perl-continued-brace-offset -4
+(defvar perl-continued-brace-offset -4
   "*Extra indent for substatements that start with open-braces.
-This is in addition to perl-continued-statement-offset.")
-(defconst perl-brace-offset 0
+This is in addition to `perl-continued-statement-offset'.")
+(defvar perl-brace-offset 0
   "*Extra indentation for braces, compared with other text in same context.")
-(defconst perl-brace-imaginary-offset 0
+(defvar perl-brace-imaginary-offset 0
   "*Imagined indentation of an open brace that actually follows a statement.")
-(defconst perl-label-offset -2
+(defvar perl-label-offset -2
   "*Offset of Perl label lines relative to usual indentation.")
 
-(defconst perl-tab-always-indent t
-  "*Non-nil means TAB in Perl mode should always indent the current line,
-regardless of where in the line point is when the TAB command is used.")
+(defvar perl-tab-always-indent t
+  "*Non-nil means TAB in Perl mode always indents the current line.
+Otherwise it inserts a tab character if you type it past the first
+nonwhite character on the line.")
 
-(defconst perl-tab-to-comment t
-  "*Non-nil means that for lines which don't need indenting, TAB will
-either indent an existing comment, move to end-of-line, or if at end-of-line
-already, create a new comment.")
+;; I changed the default to nil for consistency with general Emacs
+;; conventions -- rms.
+(defvar perl-tab-to-comment nil
+  "*Non-nil means TAB moves to eol or makes a comment in some cases.
+For lines which don't need indenting, TAB either indents an
+existing comment, moves to end-of-line, or if at end-of-line already,
+create a new comment.")
 
-(defconst perl-nochange ";?#\\|\f\\|\\s(\\|\\(\\w\\|\\s_\\)+:"
-  "*Lines starting with this regular expression will not be auto-indented.")
+(defvar perl-nochange ";?#\\|\f\\|\\s(\\|\\(\\w\\|\\s_\\)+:"
+  "*Lines starting with this regular expression are not auto-indented.")
 
+;;;###autoload
 (defun perl-mode ()
   "Major mode for editing Perl code.
 Expression and list commands understand all Perl brackets.
@@ -189,7 +214,7 @@ Variables controlling indentation style:
     either delete an empty comment, indent an existing comment, move 
     to end-of-line, or if at end-of-line already, create a new comment.
  perl-nochange
-    Lines starting with this regular expression will not be auto-indented.
+    Lines starting with this regular expression are not auto-indented.
  perl-indent-level
     Indentation of Perl statements within surrounding block.
     The surrounding block's indentation is the indentation
@@ -199,7 +224,7 @@ Variables controlling indentation style:
     then-clause of an if or body of a while.
  perl-continued-brace-offset
     Extra indentation given to a brace that starts a substatement.
-    This is in addition to perl-continued-statement-offset.
+    This is in addition to `perl-continued-statement-offset'.
  perl-brace-offset
     Extra indentation for line if it starts with an open brace.
  perl-brace-imaginary-offset
@@ -216,8 +241,7 @@ Various indentation styles:       K&R  BSD  BLK  GNU  LW
   perl-brace-imaginary-offset      0    0    4    0    0
   perl-label-offset               -5   -8   -2   -2   -2
 
-Turning on Perl mode calls the value of the variable perl-mode-hook with no 
-args, if that value is non-nil."
+Turning on Perl mode runs the normal hook `perl-mode-hook'."
   (interactive)
   (kill-all-local-variables)
   (use-local-map perl-mode-map)
@@ -226,7 +250,7 @@ args, if that value is non-nil."
   (setq local-abbrev-table perl-mode-abbrev-table)
   (set-syntax-table perl-mode-syntax-table)
   (make-local-variable 'paragraph-start)
-  (setq paragraph-start (concat "^$\\|" page-delimiter))
+  (setq paragraph-start (concat "$\\|" page-delimiter))
   (make-local-variable 'paragraph-separate)
   (setq paragraph-separate paragraph-start)
   (make-local-variable 'paragraph-ignore-fill-prefix)
@@ -246,7 +270,9 @@ args, if that value is non-nil."
   (make-local-variable 'comment-indent-function)
   (setq comment-indent-function 'perl-comment-indent)
   (make-local-variable 'parse-sexp-ignore-comments)
-  (setq parse-sexp-ignore-comments nil)
+  (setq parse-sexp-ignore-comments t)
+  (make-local-variable 'font-lock-defaults)
+  (setq font-lock-defaults '(perl-font-lock-keywords))
   (run-hooks 'perl-mode-hook))
 
 ;; This is used by indent-for-comment
@@ -257,12 +283,14 @@ args, if that value is non-nil."
       0					;Existing comment at bol stays there.
     (save-excursion
       (skip-chars-backward " \t")
-      (max (1+ (current-column))	;Else indent at comment column
-	   comment-column))))		; except leave at least one space.
+      (max (if (bolp)			;Else indent at comment column
+	       0			; except leave at least one space if
+	     (1+ (current-column)))	; not at beginning of line.
+	   comment-column))))
 
 (defun electric-perl-terminator (arg)
-  "Insert character.  If at end-of-line, and not in a comment or a quote,
-correct the line's indentation."
+  "Insert character and adjust indentation.
+If at end-of-line, and not in a comment or a quote, correct the's indentation."
   (interactive "P")
   (let ((insertpos (point)))
     (and (not arg)			; decide whether to indent
@@ -299,13 +327,13 @@ correct the line's indentation."
 
 With an argument, indent the current line, regardless of other options.
 
-If perl-tab-always-indent is nil and point is not in the indentation
+If `perl-tab-always-indent' is nil and point is not in the indentation
 area at the beginning of the line, simply insert a tab.
 
 Otherwise, indent the current line.  If point was within the indentation
 area it is moved to the end of the indentation area.  If the line was
 already indented properly and point was not within the indentation area,
-and if perl-tab-to-comment is non-nil (the default), then do the first
+and if `perl-tab-to-comment' is non-nil (the default), then do the first
 possible action from the following list:
 
   1) delete an empty comment
@@ -317,7 +345,7 @@ possible action from the following list:
   (if arg				; If arg, just indent this line
       (perl-indent-line "\f")
     (if (and (not perl-tab-always-indent)
-	     (<= (current-column) (current-indentation)))
+	     (> (current-column) (current-indentation)))
 	(insert-tab)
       (let (bof lsexp delta (oldpnt (point)))
 	(beginning-of-line) 
@@ -360,7 +388,8 @@ possible action from the following list:
 			 (ding t))))))))))))
 
 (defun perl-indent-line (&optional nochange parse-start)
-  "Indent current line as Perl code.  Return the amount the indentation 
+  "Indent current line as Perl code.
+Return the amount the indentation 
 changed by, or (parse-state) if line starts in a quoted string."
   (let ((case-fold-search nil)
 	(pos (- (point-max) (point)))
@@ -438,8 +467,8 @@ Returns (parse-state) if line starts inside a string."
 			     (memq (char-syntax (char-after (- (point) 2)))
 				   '(?w ?_))))
 	       (if (eq (preceding-char) ?\,)
-		   (perl-backward-to-start-of-continued-exp containing-sexp))
-	       (beginning-of-line)
+		   (perl-backward-to-start-of-continued-exp containing-sexp)
+		 (beginning-of-line))
 	       (perl-backward-to-noncomment))
 	     ;; Now we get the answer.
 	     (if (not (memq (preceding-char) '(?\; ?\} ?\{)))
