@@ -1,6 +1,6 @@
 ;;; scroll-bar.el --- window system-independent scroll bar support.
 
-;;; Copyright (C) 1993 Free Software Foundation, Inc.
+;;; Copyright (C) 1993, 1994, 1995 Free Software Foundation, Inc.
 
 ;; Maintainer: FSF
 ;; Keywords: hardware
@@ -34,6 +34,13 @@
 
 ;;;; Utilities.
 
+(defun scroll-bar-event-ratio (event)
+  "Given a scroll bar event EVENT, return the scroll bar position as a ratio.
+The value is a cons cell (PORTION . WHOLE) containing two integers
+whose ratio gives the event's vertical position in the scroll bar, with 0
+referring to the top and 1 to the bottom."
+  (nth 2 event))
+
 (defun scroll-bar-scale (num-denom whole)
   "Given a pair (NUM . DENOM) and WHOLE, return (/ (* NUM WHOLE) DENOM).
 This is handy for scaling a position on a scroll bar into real units,
@@ -125,18 +132,24 @@ EVENT should be a scroll bar click or drag event."
   "Scroll the window by dragging the scroll bar slider.
 If you click outside the slider, the window scrolls to bring the slider there."
   (interactive "e")
-  (let* (done)
-    (scroll-bar-drag-1 event)
-    (track-mouse
-      (while (not done)
-	(setq event (read-event))
-	(if (eq (car-safe event) 'mouse-movement)
-	    (setq event (read-event)))
-	(cond ((eq (car-safe event) 'scroll-bar-movement)
-	       (scroll-bar-drag-1 event))
-	      (t
-	       ;; Exit when we get the drag event; ignore that event.
-	       (setq done t)))))))
+  (let* (done
+	 (echo-keystrokes 0))
+    (or point-before-scroll
+	(setq point-before-scroll (point)))
+    ;; Our scrolling can move point; don't let that clear point-before-scroll.
+    (let (point-before-scroll)
+      (scroll-bar-drag-1 event)
+      (track-mouse
+	(while (not done)
+	  (setq event (read-event))
+	  (if (eq (car-safe event) 'mouse-movement)
+	      (setq event (read-event)))
+	  (cond ((eq (car-safe event) 'scroll-bar-movement)
+		 (scroll-bar-drag-1 event))
+		(t
+		 ;; Exit when we get the drag event; ignore that event.
+		 (setq done t)))))
+      (sit-for 0))))
 
 (defun scroll-bar-scroll-down (event)
   "Scroll the window's top line down to the location of the scroll bar click.
@@ -148,9 +161,13 @@ EVENT should be a scroll bar click."
 	  (let* ((end-position (event-end event))
 		 (window (nth 0 end-position))
 		 (portion-whole (nth 2 end-position)))
-	    (select-window window)
-	    (scroll-down
-	     (scroll-bar-scale portion-whole (1- (window-height))))))
+	    (let (point-before-scroll)
+	      (select-window window))
+	    (or point-before-scroll
+		(setq point-before-scroll (point)))
+	    (let (point-before-scroll)
+	      (scroll-down
+	       (scroll-bar-scale portion-whole (1- (window-height)))))))
       (select-window old-selected-window))))
 
 (defun scroll-bar-scroll-up (event)
@@ -163,9 +180,13 @@ EVENT should be a scroll bar click."
 	  (let* ((end-position (event-end event))
 		 (window (nth 0 end-position))
 		 (portion-whole (nth 2 end-position)))
-	    (select-window window)
-	    (scroll-up
-	     (scroll-bar-scale portion-whole (1- (window-height))))))
+	    (let (point-before-scroll)
+	      (select-window window))
+	    (or point-before-scroll
+		(setq point-before-scroll (point)))
+	    (let (point-before-scroll)
+	      (scroll-up
+	       (scroll-bar-scale portion-whole (1- (window-height)))))))
       (select-window old-selected-window))))
 
 
