@@ -1,11 +1,11 @@
 ;;; tpu-edt.el --- Emacs emulating TPU emulating EDT
 
-;; Copyright (C) 1993 Free Software Foundation, Inc.
+;; Copyright (C) 1993, 1994, 1995 Free Software Foundation, Inc.
 
 ;; Author: Rob Riepel <riepel@networking.stanford.edu>
 ;; Maintainer: Rob Riepel <riepel@networking.stanford.edu>
-;; Version: 3.1
-;; Keywords: edt emulations
+;; Version: 4.0
+;; Keywords: emulations
 
 ;; This file is part of GNU Emacs.
 
@@ -23,14 +23,15 @@
 ;; along with GNU Emacs; see the file COPYING.  If not, write to
 ;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
+;; TPU-edt is based on tpu.el by Jeff Kowalski and Bob Covey.
+
 ;;; Code:
 
 
 ;;;
-;;;  Revision and Version Information
+;;;  Version Information
 ;;;
-(defconst tpu-version "3.1" "TPU-edt version number.")
-(defconst tpu-revision "$Revision: 1.4 $" "Revision number of TPU-edt.")
+(defconst tpu-version "4.0" "TPU-edt version number.")
 
 
 ;;;
@@ -58,17 +59,11 @@
 ;;;     o tpu-update-mode-line  o mode line section
 ;;;
 (defconst tpu-emacs19-p (not (string-lessp emacs-version "19"))
-  "Non-NIL if we are running Lucid or GNU Emacs version 19.")
-
-(defconst tpu-gnu-emacs18-p (not tpu-emacs19-p)
-  "Non-NIL if we are running GNU Emacs version 18.")
+  "Non-nil if we are running Lucid Emacs or version 19.")
 
 (defconst tpu-lucid-emacs19-p
   (and tpu-emacs19-p (string-match "Lucid" emacs-version))
-  "Non-NIL if we are running Lucid Emacs version 19.")
-
-(defconst tpu-gnu-emacs19-p (and tpu-emacs19-p (not tpu-lucid-emacs19-p))
-  "Non-NIL if we are running GNU Emacs version 19.")
+  "Non-nil if we are running Lucid Emacs version 19.")
 
 
 ;;;
@@ -98,7 +93,7 @@ GOLD is the ASCII 7-bit escape sequence <ESC>OP.")
 
 (and tpu-lucid-emacs19-p
      (defvar minibuffer-local-ns-map (make-sparse-keymap)
-       "Hack to give Lucid emacs the same maps as GNU emacs."))
+       "Hack to give Lucid Emacs the same maps as ordinary Emacs."))
 
 
 ;;;
@@ -197,8 +192,10 @@ GOLD is the ASCII 7-bit escape sequence <ESC>OP.")
 			     (purecopy "  ")
 			     'tpu-mark-flag
 			     (purecopy " %[(")
-			     'mode-name 'minor-mode-alist "%n" 'mode-line-process
-			     (purecopy ")%]----")
+			     'mode-name 'mode-line-process 'minor-mode-alist
+			     (purecopy "%n")
+			     (purecopy ")%]--")
+			     (purecopy '(line-number-mode "L%l--"))
 			     (purecopy '(-3 . "%p"))
 			     (purecopy "-%-")))
 	 (or (assq 'tpu-newline-and-indent-p minor-mode-alist)
@@ -221,12 +218,12 @@ GOLD is the ASCII 7-bit escape sequence <ESC>OP.")
   (cond (tpu-emacs19-p (force-mode-line-update))
 	(t (set-buffer-modified-p (buffer-modified-p)) (sit-for 0))))
 
-(cond (tpu-gnu-emacs19-p
-       (add-hook 'activate-mark-hook 'tpu-update-mode-line)
-       (add-hook 'deactivate-mark-hook 'tpu-update-mode-line))
-      (tpu-lucid-emacs19-p
+(cond (tpu-lucid-emacs19-p
        (add-hook 'zmacs-deactivate-region-hook 'tpu-update-mode-line)
-       (add-hook 'zmacs-activate-region-hook 'tpu-update-mode-line)))
+       (add-hook 'zmacs-activate-region-hook 'tpu-update-mode-line))
+      (tpu-emacs19-p
+       (add-hook 'activate-mark-hook 'tpu-update-mode-line)
+       (add-hook 'deactivate-mark-hook 'tpu-update-mode-line)))
 
 
 ;;;
@@ -300,15 +297,15 @@ Otherwise sets the tpu-match markers to nil and returns nil."
 (defun tpu-mark nil
   "TPU-edt version of the mark function.
 Return the appropriate value of the mark for the current
-version of emacs."
+version of Emacs."
   (cond (tpu-lucid-emacs19-p (mark (not zmacs-regions)))
-	(tpu-gnu-emacs19-p (and mark-active (mark (not transient-mark-mode))))
+	(tpu-emacs19-p (and mark-active (mark (not transient-mark-mode))))
 	(t (mark))))
 
 (defun tpu-set-mark (pos)
-  "TPU-edt verion of the set-mark function.
+  "TPU-edt verion of the `set-mark' function.
 Sets the mark at POS and activates the region acording to the
-current version of emacs."
+current version of Emacs."
   (set-mark pos)
   (and tpu-lucid-emacs19-p pos (zmacs-activate-region)))
 
@@ -323,7 +320,7 @@ current version of emacs."
 (defun tpu-y-or-n-p (prompt &optional not-yes)
   "Prompt for a y or n answer with positive default.
 Optional second argument NOT-YES changes default to negative.
-Like emacs y-or-n-p, also accepts space as y and DEL as n."
+Like Emacs `y-or-n-p', but also accepts space as y and DEL as n."
   (message (format "%s[%s]" prompt (if not-yes "n" "y")))
   (let ((doit t))
     (while doit
@@ -477,6 +474,14 @@ its ASCII decimal value."
   (if overwrite-mode (delete-char 1))
   (insert (if num num 0)))
 
+(defun tpu-quoted-insert (num)
+  "Read next input character and insert it.
+This is useful for inserting control characters."
+  (interactive "*p")
+  (let ((char (read-char)) )
+    (if overwrite-mode (delete-char num))
+    (insert-char char num)))
+
 
 ;;;
 ;;;  TPU line-mode commands
@@ -561,6 +566,16 @@ its ASCII decimal value."
 (fset 'help 'tpu-help)
 (fset 'HELP 'tpu-help)
 
+(fset 'set\ cursor\ free 'tpu-set-cursor-free)
+(fset 'SET\ CURSOR\ FREE 'tpu-set-cursor-free)
+
+(fset 'set\ cursor\ bound 'tpu-set-cursor-bound)
+(fset 'SET\ CURSOR\ BOUND 'tpu-set-cursor-bound)
+
+(fset 'set\ scroll\ margins 'tpu-set-scroll-margins)
+(fset 'SET\ SCROLL\ MARGINS 'tpu-set-scroll-margins)
+
+
 ;; Around emacs version 18.57, function line-move was renamed to
 ;; next-line-internal.  If we're running under an older emacs,
 ;; make next-line-internal equivalent to line-move.
@@ -768,8 +783,8 @@ kills modified buffers without asking."
 (defun tpu-save-all-buffers-kill-emacs nil
   "Save all buffers and exit emacs."
   (interactive)
-  (setq trim-versions-without-asking t)
-  (save-buffers-kill-emacs t))
+  (let ((delete-old-versions t))
+    (save-buffers-kill-emacs t)))
 
 (defun tpu-write-current-buffers nil
   "Save all modified buffers without exiting."
@@ -782,14 +797,19 @@ kills modified buffers without asking."
   (switch-to-buffer (car (reverse (buffer-list)))))
 
 (defun tpu-next-file-buffer nil
-  "Go to next buffer in ring that is visiting a file."
+  "Go to next buffer in ring that is visiting a file or directory."
   (interactive)
-  (let ((starting-buffer (buffer-name)))
-    (switch-to-buffer (car (reverse (buffer-list))))
-    (while (and (not (equal (buffer-name) starting-buffer))
-		(not (buffer-file-name)))
-      (switch-to-buffer (car (reverse (buffer-list)))))
-    (if (equal (buffer-name) starting-buffer) (error "No other buffers."))))
+  (let ((list (tpu-make-file-buffer-list (buffer-list))))
+    (setq list (delq (current-buffer) list))
+    (if (not list) (error "No other buffers."))
+    (switch-to-buffer (car (reverse list)))))
+
+(defun tpu-make-file-buffer-list (buffer-list)
+  "Returns names from BUFFER-LIST excluding those beginning with a space or star."
+  (delq nil (mapcar '(lambda (b)
+                       (if (or (= (aref (buffer-name b) 0) ? )
+                               (= (aref (buffer-name b) 0) ?*)) nil b))
+                    buffer-list)))
 
 (defun tpu-next-window nil
   "Move to the next window."
@@ -858,10 +878,11 @@ The search is performed in the current direction."
 ;;  to ensure that the next search will be in the current direction.  It is
 ;;  called from:
 
-;;       tpu-advance              tpu-backup
-;;       tpu-toggle-regexp        tpu-toggle-search-direction (t)
-;;       tpu-search               tpu-lm-replace
-;;       tpu-search-forward (t)   tpu-search-reverse (t)
+;;       tpu-advance                   tpu-backup
+;;       tpu-toggle-regexp             tpu-toggle-search-direction (t)
+;;       tpu-search                    tpu-lm-replace
+;;       tpu-search-forward (t)        tpu-search-reverse (t)
+;;       tpu-search-forward-exit (t)   tpu-search-backward-exit (t)
 
 (defun tpu-set-search (&optional arg)
   "Set the search functions and set the search direction to the current
@@ -890,33 +911,50 @@ direction.  If an argument is specified, don't set the search direction."
   (tpu-unset-match)
   (tpu-adjust-search)
 
-  (cond ((tpu-emacs-search tpu-search-last-string nil t)
-	 (tpu-set-match) (goto-char (tpu-match-beginning)))
+  (let ((case-fold-search
+	 (and case-fold-search (tpu-check-search-case tpu-search-last-string))))
 
-	(t
-	 (tpu-adjust-search t)
-	 (let ((found nil) (pos nil))
-	   (save-excursion
-	     (let ((tpu-searching-forward (not tpu-searching-forward)))
-	       (tpu-adjust-search)
-	       (setq found (tpu-emacs-rev-search tpu-search-last-string nil t))
-	       (setq pos (match-beginning 0))))
+    (cond ((tpu-emacs-search tpu-search-last-string nil t)
+	   (tpu-set-match) (goto-char (tpu-match-beginning)))
 
-	   (cond (found
-		  (cond ((tpu-y-or-n-p
-			  (format "Found in %s direction.  Go there? "
-				  (if tpu-searching-forward "reverse" "forward")))
-			 (goto-char pos) (tpu-set-match)
-			 (tpu-toggle-search-direction))))
+	  (t
+	   (tpu-adjust-search t)
+	   (let ((found nil) (pos nil))
+	     (save-excursion
+	       (let ((tpu-searching-forward (not tpu-searching-forward)))
+		 (tpu-adjust-search)
+		 (setq found (tpu-emacs-rev-search tpu-search-last-string nil t))
+		 (setq pos (match-beginning 0))))
 
-		 (t
-		  (if (not quiet)
-		      (message
-		       "%sSearch failed: \"%s\""
-		       (if tpu-regexp-p "RE " "") tpu-search-last-string))))))))
+	     (cond
+	      (found
+	       (cond ((tpu-y-or-n-p
+		       (format "Found in %s direction.  Go there? "
+			       (if tpu-searching-forward "reverse" "forward")))
+		      (goto-char pos) (tpu-set-match)
+		      (tpu-toggle-search-direction))))
+
+	      (t
+	       (if (not quiet)
+		   (message
+		    "%sSearch failed: \"%s\""
+		    (if tpu-regexp-p "RE " "") tpu-search-last-string)))))))))
 
 (fset 'tpu-search-internal-core (symbol-function 'tpu-search-internal))
 
+(defun tpu-check-search-case (string)
+  "Returns t if string contains upper case."
+  ;; if using regexp, elimiate upper case forms (\B \W \S.)
+  (if tpu-regexp-p
+      (let ((pat (copy-sequence string)) (case-fold-search nil) (pos 0))
+	(while (setq pos (string-match "\\\\\\\\" pat)) (aset pat (+ 1 pos) ?.))
+	(while (setq pos (string-match "\\\\B" pat)) (aset pat (+ 1 pos) ?.))
+	(while (setq pos (string-match "\\\\W" pat)) (aset pat (+ 1 pos) ?.))
+	(while (setq pos (string-match "\\\\S." pat))
+	  (aset pat (+ 1 pos) ?.) (aset pat (+ 2 pos) ?.))
+	(string-equal pat (downcase pat)))
+    (string-equal string (downcase string))))
+
 (defun tpu-adjust-search (&optional arg)
   "For forward searches, move forward a character before searching,
 and backward a character after a failed search.  Arg means end of search."
@@ -934,6 +972,20 @@ Used for reversing a search in progress."
        (message "Searching %sward."
 		(if tpu-searching-forward "for" "back"))))
 
+(defun tpu-search-forward-exit nil
+  "Set search direction forward and exit minibuffer."
+  (interactive)
+  (setq tpu-searching-forward t)
+  (tpu-set-search t)
+  (exit-minibuffer))
+
+(defun tpu-search-backward-exit nil
+  "Set search direction backward and exit minibuffer."
+  (interactive)
+  (setq tpu-searching-forward nil)
+  (tpu-set-search t)
+  (exit-minibuffer))
+
 
 ;;;
 ;;;  Select / Unselect
@@ -976,7 +1028,7 @@ corners of a rectangle."
   (let ((mc (current-column))
 	(pc (progn (exchange-point-and-mark) (current-column))))
 
-    (cond ((> (point) (tpu-mark))                      ; point on lower line
+    (cond ((> (point) (tpu-mark))                  ; point on lower line
 	   (cond ((> pc mc)                        ; point @  lower-right
 		  (exchange-point-and-mark))       ; point -> upper-left
 
@@ -1798,7 +1850,7 @@ Accepts a prefix argument for the number of tpu-pan-columns to scroll."
 (define-key GOLD-map "C" 'repeat-complex-command)             ; C
 (define-key GOLD-map "D" 'shell-command)                      ; D
 (define-key GOLD-map "E" 'tpu-exit)                           ; E
-(define-key GOLD-map "F" 'nil)                                ; F
+(define-key GOLD-map "F" 'tpu-set-cursor-free)                ; F
 (define-key GOLD-map "G" 'tpu-get)                            ; G
 (define-key GOLD-map "H" 'nil)                                ; H
 (define-key GOLD-map "I" 'tpu-include)                        ; I
@@ -1829,7 +1881,7 @@ Accepts a prefix argument for the number of tpu-pan-columns to scroll."
 (define-key GOLD-map "c" 'repeat-complex-command)             ; c
 (define-key GOLD-map "d" 'shell-command)                      ; d
 (define-key GOLD-map "e" 'tpu-exit)                           ; e
-(define-key GOLD-map "f" 'nil)                                ; f
+(define-key GOLD-map "f" 'tpu-set-cursor-free)                ; f
 (define-key GOLD-map "g" 'tpu-get)                            ; g
 (define-key GOLD-map "h" 'nil)                                ; h
 (define-key GOLD-map "i" 'tpu-include)                        ; i
@@ -1942,6 +1994,13 @@ Accepts a prefix argument for the number of tpu-pan-columns to scroll."
      (define-key repeat-complex-command-map "\eOM" 'exit-minibuffer))
 
 
+;;;
+;;;  Minibuffer map additions to set search direction
+;;;
+(define-key minibuffer-local-map "\eOt" 'tpu-search-forward-exit)
+(define-key minibuffer-local-map "\eOu" 'tpu-search-backward-exit)
+
+
 ;;;
 ;;;  Map control keys
 ;;;
@@ -1949,14 +2008,13 @@ Accepts a prefix argument for the number of tpu-pan-columns to scroll."
 (define-key global-map "\C-a" 'tpu-toggle-overwrite-mode)     ; ^A
 (define-key global-map "\C-b" 'repeat-complex-command)        ; ^B
 (define-key global-map "\C-e" 'tpu-current-end-of-line)       ; ^E
-(define-key global-map "\C-f" 'set-visited-file-name)         ; ^F
 (define-key global-map "\C-h" 'tpu-next-beginning-of-line)    ; ^H (BS)
 (define-key global-map "\C-j" 'tpu-delete-previous-word)      ; ^J (LF)
 (define-key global-map "\C-k" 'tpu-define-macro-key)          ; ^K
 (define-key global-map "\C-l" 'tpu-insert-formfeed)           ; ^L (FF)
 (define-key global-map "\C-r" 'recenter)                      ; ^R
 (define-key global-map "\C-u" 'tpu-delete-to-bol)             ; ^U
-(define-key global-map "\C-v" 'quoted-insert)                 ; ^V
+(define-key global-map "\C-v" 'tpu-quoted-insert)             ; ^V
 (define-key global-map "\C-w" 'redraw-display)                ; ^W
 (define-key global-map "\C-z" 'tpu-exit)                      ; ^Z
 
@@ -1979,7 +2037,6 @@ Accepts a prefix argument for the number of tpu-pan-columns to scroll."
  	     (define-key global-map "\C-a" (lookup-key map "\C-a"))     ; ^A
  	     (define-key global-map "\C-b" (lookup-key map "\C-b"))     ; ^B
  	     (define-key global-map "\C-e" (lookup-key map "\C-e"))     ; ^E
- 	     (define-key global-map "\C-f" (lookup-key map "\C-f"))     ; ^F
 	     (define-key global-map "\C-h" (lookup-key map "\C-h"))     ; ^H (BS)
 	     (define-key global-map "\C-j" (lookup-key map "\C-j"))     ; ^J (LF)
 	     (define-key global-map "\C-k" (lookup-key map "\C-k"))     ; ^K
@@ -2043,16 +2100,19 @@ Accepts a prefix argument for the number of tpu-pan-columns to scroll."
 (defun tpu-load-xkeys (file)
   "Load the TPU-edt X-windows key definitions FILE.
 If FILE is nil, try to load a default file.  The default file names are
-~/.tpu-lucid-keys for Lucid emacs, and ~/.tpu-gnu-keys for GNU emacs."
+`~/.tpu-lucid-keys' for Lucid emacs, and `~/.tpu-keys' for Emacs."
   (interactive "fX key definition file: ")
   (cond (file
 	 (setq file (expand-file-name file)))
 	(tpu-xkeys-file
 	 (setq file (expand-file-name tpu-xkeys-file)))
-	(tpu-gnu-emacs19-p
-	 (setq file (expand-file-name "~/.tpu-gnu-keys")))
 	(tpu-lucid-emacs19-p
-	 (setq file (expand-file-name "~/.tpu-lucid-keys"))))
+	 (setq file (expand-file-name "~/.tpu-lucid-keys")))
+	(tpu-emacs19-p
+	 (setq file (expand-file-name "~/.tpu-keys"))
+	 (and (not (file-exists-p file))
+	      (file-exists-p (expand-file-name "~/.tpu-gnu-keys"))
+	      (tpu-copy-keyfile (expand-file-name "~/.tpu-gnu-keys") file))))
   (cond ((file-readable-p file)
 	 (load-file file))
 	(t
@@ -2085,6 +2145,34 @@ If FILE is nil, try to load a default file.  The default file names are
 		  (insert "Nope, I can't seem to find it.  :-(\n\n")
 		  (sit-for 120)))))))
 
+(defun tpu-copy-keyfile (oldname newname)
+  "Copy the TPU-edt X key definitions file to the new default name."
+  (interactive "fOld name: \nFNew name: ")
+  (if (not (get-buffer "*TPU-Notice*")) (generate-new-buffer "*TPU-Notice*"))
+  (set-buffer "*TPU-Notice*")
+  (erase-buffer)
+  (insert "
+  NOTICE --
+
+  The default name of the TPU-edt key definition file has changed
+  from `~/.tpu-gnu-keys' to `~/.tpu-keys'.  With your permission,
+  your key definitions will be copied to the new file.  If you'll
+  never use older versions of Emacs, you can remove the old file.
+  If the copy fails, you'll be asked if you want to create a new
+  key definitions file.  Do you want to copy your key definition
+  file now?
+  ")
+  (save-window-excursion
+    (switch-to-buffer-other-window "*TPU-Notice*")
+    (shrink-window-if-larger-than-buffer)
+    (goto-char (point-min))
+    (beep)
+    (and (tpu-y-or-n-p "Copy key definitions to the new file now? ")
+	 (condition-case conditions
+             (copy-file oldname newname)
+	   (error (message "Sorry, couldn't copy - %s" (cdr conditions)))))
+    (kill-buffer "*TPU-Notice*")))
+
 
 ;;;
 ;;;  Start and Stop TPU-edt
@@ -2125,15 +2213,10 @@ If FILE is nil, try to load a default file.  The default file names are
     (setq-default page-delimiter "^\f")
     (setq-default truncate-lines nil)
     (setq scroll-step 0)
+    (setq global-map (copy-keymap tpu-original-global-map))
     (use-global-map global-map)
     (setq tpu-edt-mode nil))))
 
-
-;;;
-;;;  Turn on TPU-edt and announce it as a feature
-;;;
-(tpu-edt-mode)
-
 (provide 'tpu-edt)
 
 ;;; tpu-edt.el ends here
