@@ -158,10 +158,6 @@ No message."
     (forward-char (1- val))
     (if com (vip-execute-com 'vip-goto-col val com))))
 
-(defun vip-refresh-mode-line ()
-  "Redraw mode line."
-  (set-buffer-modified-p (buffer-modified-p)))
-
 (defun vip-copy-keymap (map)
   (if (null map) (make-sparse-keymap) (copy-keymap map)))
 
@@ -207,7 +203,7 @@ No message."
 	       (vip-change-mode-line "Emacs:")
 	       (use-local-map vip-emacs-local-map)))
 	(setq vip-current-mode new-mode)
-	(vip-refresh-mode-line))))
+	(force-mode-line-update))))
 
 (defun vip-copy-region-as-kill (beg end)
   "If BEG and END do not belong to the same buffer, it copies empty region."
@@ -627,7 +623,7 @@ to vip-d-com for later use by vip-repeat"
 
 (defun vip-repeat (arg)
   "(ARG)  Re-execute last destructive command.  vip-d-com has the form
-(COM ARG CH REG), where COM is the command to be re-executed, ARG is the
+\(COM ARG CH REG), where COM is the command to be re-executed, ARG is the
 argument for COM, CH is a flag for repeat, and REG is optional and if exists
 is the name of the register for COM."
   (interactive "P")
