@@ -1,6 +1,6 @@
 /* Hooks by which low level terminal operations
    can be made to call other routines.
-   Copyright (C) 1985, 1986, 1993 Free Software Foundation, Inc.
+   Copyright (C) 1985, 1986, 1993, 1994 Free Software Foundation, Inc.
 
 This file is part of GNU Emacs.
 
@@ -183,10 +183,14 @@ extern void (*judge_scroll_bars_hook)( /* FRAME_PTR *FRAME */ );
 
 /* Input queue declarations and hooks.  */
 
+/* Called to read input events.  */
 extern int (*read_socket_hook) ();
 
+/* Called when a frame's display becomes entirely up to date.  */
+extern int (*frame_up_to_date_hook) ();
+
 /* Expedient hack: only provide the below definitions to files that
-   are prepared to handle lispy things.  XINT is defined iff lisp.h
+   are prepared to handle lispy things.  CONSP is defined iff lisp.h
    has been included before this file.  */
 #ifdef CONSP
 
@@ -241,7 +245,15 @@ enum event_kind
 				   milliseconds) for the click.  */
   selection_request_event,	/* Another X client wants a selection from us.
 				   See `struct selection_event'.  */
-  selection_clear_event		/* Another X client cleared our selection.  */
+  selection_clear_event,	/* Another X client cleared our selection.  */
+  buffer_switch_event,		/* A process filter has switched buffers.  */
+  delete_window_event,		/* An X client said "delete this window".  */
+  menu_bar_event,		/* An event generated by the menu bar.
+				   The frame_or_window field's cdr holds the
+				   Lisp-level event value.
+				   (Only the toolkit version uses these.)  */
+  iconify_event,		/* An X client iconified this window.  */
+  deiconify_event		/* An X client deiconified this window.  */
 };
 
 /* If a struct input_event has a kind which is selection_request_event
@@ -258,11 +270,16 @@ struct input_event {
   /* What kind of event was this?  */
   enum event_kind kind;
   
-  Lisp_Object code;
+  /* For an ascii_keystroke, this is the character.
+     For a non_ascii_keystroke, this is the keysym code.
+     For a mouse event, this is the button number.  */
+  int code;
   enum scroll_bar_part part;
 
   /* This field is copied into a vector while the event is in the queue,
      so that garbage collections won't kill it.  */
+  /* In a menu_bar_event, this is a cons cell whose car is the frame
+     and whose cdr is the Lisp object that is the event's value.  */
   Lisp_Object frame_or_window;
 
   int modifiers;		/* See enum below for interpretation.  */
